/*
 * CSE3001.  Compiler project based on "Modern Compiler Implementation
 * in Java" by Andrew W. Appel.
 *
 */

package error;

import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;

/**
 ** A class to maintain line numbers for printing error messages.
 **
 **/

public class ErrorMsg {
 
   private LinkedList posList = new LinkedList ();
   private int lineNum=1;      // first line of file is line number 1
   private String filename;
   private int errors = 0;
   public int getErrors () { return errors; }

   public ErrorMsg(String f) {
      filename=f;
   }

   // Character position of each line needs to be recorded
   public void newline (int pos) {
      lineNum++;
      posList.addFirst (new Integer(pos));
   }

   /*
     Given the position of the error, the line number of the file can
     be recovered by searching backward to some line that ends with a small
     character position.   We search backward because it seems more likely
     that error will be detected in the last few lines.
   */
   private String position (int pos) {
      int n = lineNum;
      for (Iterator i = posList.iterator(); i.hasNext(); n--) {
	 final int m = ((Integer) i.next()).intValue();
	 if (m<pos) return String.valueOf(n) + "." + String.valueOf (pos-m);
      }
      return "1." + String.valueOf (pos);
   }

   public void error (int pos, String msg) {
      errors++;
      System.err.println(filename + ":" + position (pos) + ": " + msg);
   }

   // A "main" for unit testing
   public static void main (String args []) {
      final ErrorMsg e = new ErrorMsg ("test");
      for (int i=0; i<args.length-1; i++) {
	 e.newline (Integer.parseInt (args[i]));
      }
      System.out.println (e.position (Integer.parseInt (args[args.length-1])));
   }

}
