package regalloc;

import java.util.HashMap;

import tree.*;

public class Cost 
{
    public void put(Object o, int i) { hm.put(o, new Integer(i)); }
    public int  get(Object o) { return ((Integer)hm.get(o)).intValue(); }

    HashMap hm;
    public Cost(StmList stms)
    {
        hm = new HashMap();
        for(StmList sl = stms; sl!=null; sl=sl.tail)
            label(sl.head);
    }

    private void label(Stm s) 
    {
        if(s instanceof CJUMP)      label((CJUMP)s);
        else if(s instanceof EXP)   label((EXP)s);
        else if(s instanceof JUMP)  label((JUMP)s);
        else if(s instanceof LABEL) label((LABEL)s);
        else if(s instanceof MOVE)  label((MOVE)s);
        else if(s instanceof NOP)   label((NOP)s);
        else if(s instanceof SEQ)   label((SEQ)s);
        else throw new Error("Unknown class: " + s.getClass());
    }

    private void label(NOP n)   { solve(n, 0); }
    private void label(JUMP j)  { solve(j, 0); }
    private void label(LABEL l) { solve(l, 0); }

    private void label(CJUMP s) 
    {
        label(s.left);
        label(s.right);
        solve(s, s.left, s.right);
    }

    private void label(EXP e)   {
        label(e.exp);
        solve(e, e.exp);
    }

    private void label(MOVE m) 
    {
        label(m.dst);
        label(m.src);
        solve(m, m.dst, m.src);
    }

    private void label(SEQ s)
    {
        label(s.left);
        label(s.right);
        solve(s, s.left, s.right);
        solve(s, 0);
    }
    
    private void label(Exp s) 
    {
        if(s instanceof BINOP)      label((BINOP)s);
        else if(s instanceof CALL)  label((CALL)s);
        else if(s instanceof CONST) label((CONST)s);
        else if(s instanceof TEMP)  label((TEMP)s); 
        else if(s instanceof MEM)   label((MEM)s);
        else if(s instanceof NAME)  label((NAME)s); 
        else if(s instanceof ESEQ)  label((ESEQ)s);
        else throw new Error("Unknown class: " + s.getClass());
    }

    private void label(CONST c) { solve(c, 1); }
    private void label(TEMP t)  { solve(t, 0); }
    private void label(NAME n)  { solve(n, 1); }

    private void label(BINOP b) 
    {
        label(b.left);
        label(b.right);
        solve(b, b.left, b.right);
    }

    private void label(CALL c) 
    {
        label(c.func);
        label(c.args);
        solve(c, c.func, c.args);
    }

    private void label(ExpList e)
    {
        for(ExpList l=e; l!=null; l=l.tail)
            label(l.head);
        solve(e, 0);
    }
    
    private void label(MEM m)
    {
        label(m.exp);
        solve(m, m.exp);
    }

    private void label(ESEQ e) 
    {
        label(e.stm);
        label(e.exp);
        solve(e, e.stm, e.exp);
    }

    ///////////////////////////////////////
    
    private final void solve(Object o, int c)
    { put(o, c); }
    
    private final void solve(Object o, Object ca)
    { put(o, cost(ca)); }

    private final void solve(Object o, Object ca, Object cb)
    { put(o, cost(ca, cb)); }

    private final int cost(Object a)           { return max(1, get(a));       }
    private final int cost(Object a, Object b) { return cost(get(a), get(b)); }
    private final int cost(int a, int b)  { return a==b ? a+1 : max(1, a, b); }

    private final int max(int a, int b, int c)   { return max(a, max(b, c)); }
    private final int max(int a, int b)          { return (a>b ? a : b);     }
    private final int min(int a, int b)          { return (a<b ? a : b);     }
}
