package sparc;

import tree.*;

import temp.Label;
import temp.Temp;
import temp.LabelList;
import temp.TempList;

import java.util.HashMap;
import java.util.List;
import java.util.LinkedList;

import tree.ExpList;

import assem.Instr;

import frame.Frame;

import regalloc.RegAlloc;

public class Codegen 
{
    private static final Registers registers = SparcFrame.registers;
    
    private void DIE(String s) 
    { System.out.println("FATAL ERROR: " + s); System.exit(1); }
    
    private assem.OPER oper(String a, TempList d, TempList s, LabelList j)
    { return new assem.OPER("\t" + a,
                            d!=null ? d.toList() : null,
                            s!=null ? s.toList() : null,
                            j!=null ? j.toList() : null); }

    private assem.OPER oper(String a, TempList d, TempList s)
    { return oper(a, d, s, null); }
                            
    TempList L(Temp t, TempList l) { return new TempList(t, l); }
    TempList L(Temp a)             { return L(a, (TempList)null); }
    TempList L(Temp a, Temp b)     { return L(a, L(b)); }

    LabelList L(Label a, LabelList l) { return new LabelList(a, l); }
    LabelList L(Label a)              { return L(a, (LabelList)null); }
    LabelList L(Label a, Label b)     { return L(a, L(b)); }
    
    List ilist;

    private void emit(Instr inst) 
    { ilist.add(inst); }

    private Temp getReg(Object o)
    {
        if(ra.get(o)==null) throw new Error("Unknown temp"); else
        return ra.get(o);
    }

    private Temp tempReg()
    { return registers.regMap("%i0"); }
    
    RegAlloc ra;
    /*
      generate instruction for statements
      return list of instructions
    */
    public List codegen(StmList stms, RegAlloc ra) 
    {
        this.ra = ra;
        
        ilist = new LinkedList();
        
        for(StmList sl = stms; sl!=null; sl=sl.tail)
            munchStm(sl.head);

        return ilist;
    }

    // pacman time!
    private void munchStm(Stm s) {
        if(s instanceof MOVE)       munchMove(((MOVE)s).dst,((MOVE)s).src);
        else if(s instanceof EXP)   munchExp(((EXP)s).exp);
        else if(s instanceof JUMP)  munchJump((JUMP)s);
        else if(s instanceof CJUMP) munchCJump((CJUMP)s);
        else if(s instanceof SEQ)   munchSeq((SEQ)s );
        else if(s instanceof LABEL) munchLabel((LABEL)s );
        else if(s instanceof NOP)   return;
        else DIE("Unknown type: " + s.getClass() + " [munchStm]");
    }
    
    void munchMove(Exp dst, Exp src) 
    {
        if(dst instanceof MEM)        munchMove((MEM)dst, src);
        else if(dst instanceof TEMP)  munchMove((TEMP)dst, src);
        else DIE("Unknown type: " + dst.getClass() + " [munchMove]");
    }
    
    void munchMove(MEM dst, Exp src) 
    {
        if(src instanceof CONST) {
            Temp r = tempReg();
            munchMove(r, src);
            munchMove(dst, r);
        }
        // MOVE (sec + offset) || (offset + src) <-- src
        else if((dst.exp instanceof BINOP) &&
                (((BINOP)dst.exp).binop==BINOP.PLUS) &&
                ( (((BINOP)dst.exp).left instanceof CONST) ||
                  (((BINOP)dst.exp).right instanceof CONST) )) {
            BINOP bne = (BINOP)dst.exp;
            if(bne.left instanceof CONST ) // offset + sec
                munchMove(bne.right, (CONST)bne.left, src);
            else
                munchMove(bne.left, (CONST)bne.right, src);
        }
        // MOVE MEM <-- MEM
        else if(src instanceof MEM)
            munchMove(dst, (MEM)src);
        else 
            emit(oper("st \t`s0, [`d0]",
                      L(munchExp(dst.exp)), L(munchExp(src))));
    }

    // MOVE section + const , src
    private void munchMove(Exp dst, CONST offset, Exp src) 
    {
        emit(oper("st \t`s0, [`d0 + " + offset.value + "]",
                  L(munchExp(dst)), L(munchExp(src))));
    }

    // MOVE mem, mem    
    private void munchMove(MEM d, MEM s)
    {
        // temporary register
        Temp r = getReg(s);
        emit(oper("mov \t`s0, `d0", L(r), L(munchExp(s))));
        munchMove(d, r);
    }

    private void munchMove(Temp dst, Temp src)
    { munchMove(new TEMP(dst), new TEMP(src)); }
    
    private void munchMove(Temp dst, Exp src)
    { munchMove(new TEMP(dst), src); }

    private void munchMove(Exp dst, Temp src)
    { munchMove(dst, new TEMP(src)); }
    
    // MOVE temp, src
    private void munchMove(TEMP dst, Exp src) 
    {
        if(!mapped(dst.temp))
            mapTemp(dst.temp, getReg(src));

        if(src instanceof NAME)       munchMove(dst, (NAME)src);
        else if(src instanceof CONST) munchMove(dst, (CONST)src);
        else if(src instanceof MEM)   munchMove(dst, (MEM)src);
        else {
            Temp rd = dst.temp;
            Temp rs = munchExp(src);
            if(!mapped(rd))
                mapTemp(rd, rs);
            munchMove(rd, rs);
        }
    }

    private void munchMove(TEMP dst, TEMP src) 
    {
        if(!equal(dst.temp, src.temp))
            emit(oper("mov \t`s0, `d0" , L(dst.temp), L(src.temp)));
    }
    
    private void munchMove(TEMP dst, CONST src) 
    { emit(oper("mov \t" + src.value + ", `d0", L(dst.temp), null)); }
    
    private void munchMove(TEMP dst, NAME src) 
    { emit(oper("set \t" + src.label + ", `d0", L(dst.temp), null)); }

    private void munchMove(TEMP dst, MEM src) 
    {
        Temp dest = dst.temp;
        // MEM(c) - move const into reg
        if(src.exp instanceof CONST)
            emit(oper("ld \t[" +  ((CONST)src.exp).value + "], `d0",
                      L(dest), null));
        //
        else if(src.exp instanceof BINOP
                && ((BINOP)src.exp).binop==BINOP.PLUS
                && ((BINOP)src.exp).left instanceof CONST) {
            // MEM(const  + exp))
            CONST left = (CONST)(((BINOP)src.exp).left);
            Exp   right = ((BINOP)src.exp).right;
            emit(oper("ld \t[`s0 + " + left.value + "], `d0",
                      L(dest), L(munchExp(right))));
        }
        else if(src.exp instanceof BINOP
                && ((BINOP)src.exp).binop==BINOP.PLUS
                && ((BINOP)src.exp).right instanceof CONST) {
            // MEM(exp + const)
            Exp   left = ((BINOP)src.exp).left;
            CONST right = (CONST)(((BINOP)src.exp).right);
            emit(oper("ld \t[`s0 + " + right.value + "], `d0",
                      L(dest), L(munchExp(left))));
        }
        else {
            //MEM(e1)
            emit(oper("ld \t[`s0], `d0", L(dest), L(munchExp(src.exp))));
        }
    }
    
    private void munchJump(JUMP j) 
    {
        if(j.targets.head!=null) {
            emit(oper("ba \t" + j.targets.head, null, null, j.targets));
            emit(oper("nop",null,null));
        }
    }

    private void munchCJump(CJUMP j) 
    {
        String instr;
        switch(j.relop) {
        case CJUMP.EQ:  instr = "be"; break;
        case CJUMP.NE:  instr = "bne"; break;
        case CJUMP.LT:  instr = "bl"; break;
        case CJUMP.GT:  instr = "bg"; break;
        case CJUMP.LE:  instr = "ble"; break;
        case CJUMP.GE:  instr = "bge"; break;
        case CJUMP.ULT: instr = "blu"; break;
        case CJUMP.ULE: instr = "bleu"; break;
        case CJUMP.UGT: instr = "bgu"; break;
        case CJUMP.UGE: instr = "bgeu"; break;
        default:
            throw new Error("Unknown CJUMP operation");
        }

        if(j.right instanceof CONST)
            emit(oper("cmp \t`s0, " + ((CONST)j.right).value, null,
                 L(munchExp(j.left))));
        else
            emit(oper("cmp \t`s0,`s1", null,
                  L(munchExp(j.left),munchExp(j.right))));
        emit(oper(instr + "\t`j0", null, null, L(j.iftrue)));
    }
    
    private void munchSeq(SEQ s) {
        munchStm(s.left);
        munchStm(s.right);
    }

    private void munchLabel(LABEL label) 
    { emit(new assem.LABEL(label.label.toString() + ":", label.label)); }

    private Temp munchExp(Exp e) {
        if(e instanceof MEM)   return munchExp((MEM)e);
        if(e instanceof BINOP) return munchExp((BINOP)e);
        if(e instanceof CONST) return munchExp((CONST)e);
        if(e instanceof TEMP)  return munchExp((TEMP)e);
        if(e instanceof CALL)  return munchExp((CALL)e);
        if(e instanceof NAME)  return munchExp((NAME)e);
        DIE("Unknown type: " + e.getClass() + " [munchExp]");
        return null;
    }

    private Temp munchExp(MEM m) {
        Temp r = getReg(m); // return value
        munchMove(r, m);
        return r;
    }

    private Temp munchExp(BINOP b) {
        if(b.binop==BINOP.MUL || b.binop==BINOP.DIV) {
            String instr = (b.binop==BINOP.MUL ? ".mul" : ".div");
            RegList regs = SparcFrame.outRegs();
            munchMove(regs.get(), b.left);
            munchMove(regs.get(), b.right);
            emit(oper("call \t" + instr, null, null));
            emit(oper("nop", null, null));
            return SparcFrame.getRV();
        }
        else {
            Temp rd = getReg(b);
            
            String instr = null;
            
            switch(b.binop){
            case BINOP.PLUS:    instr = "add"; break;
            case BINOP.MINUS:   instr = "sub"; break;
            case BINOP.AND:     instr = "and"; break;
            case BINOP.OR:      instr = "or"; break;
            case BINOP.LSHIFT:  instr = "sll"; break;
            case BINOP.RSHIFT:  instr = "srl"; break;
            case BINOP.ARSHIFT: instr = "sra"; break;
            case BINOP.XOR:     instr = "xor"; break;
            default: DIE("Unknown BINOP.binop");
            }

            boolean lc = b.left instanceof CONST;
            boolean rc = b.right instanceof CONST;
            
            if(lc || rc) {
                int c;
                Temp rs;
                
                if(lc) {
                    c = ((CONST)b.left).value;
                    rs = munchExp(b.right);
                }
                else {
                    c = ((CONST)b.right).value;
                    rs = munchExp(b.left);
                }
            
                emit(oper(instr + "\t`s0, " + c + ", `d0", L(rd), L(rs)));
            }
            else
                emit(oper(instr + "\t`s0, `s1, `d0",
                  L(rd), L(munchExp(b.left), munchExp(b.right))));
            return rd;
        }
    }
    
    private Temp munchExp(CONST c) {
        Temp r = getReg(c);
        munchMove(r, c);
        return r;
    }

    private Temp munchExp(TEMP t) {
        return t.temp;
    }

    private Temp munchExp(CALL c) {

        Temp r = SparcFrame.getRV();
        TempList args = null;
        
        if(c.func instanceof NAME) {
            RegList regs = SparcFrame.outRegs();
            for(ExpList curr = c.args; curr!=null; curr=curr.tail) {
                Temp t = regs.get();
                munchMove(t, curr.head);
                args = L(t, args);    
            }
            emit(oper("call \t" + ((NAME)c.func).label, L(r), null));
            emit(oper("nop", null, null));
        }
        return r;
    }

    private Temp munchExp(NAME t) {
        Temp r = getReg(t);
        emit(oper("set \t" + t.label + ", `d0", L(r), null));
        return r;
    }

    private boolean mapped(Temp temp)
    { return registers.tempMap(temp)!=null; }

    private void mapTemp(Temp a, Temp b)
    { registers.mapTemp(a, b); }

    private boolean equal(Temp a, Temp b) 
    { return registers.tempMap(a)==registers.tempMap(b); }
}
