package sparc;

import assem.Proc;

import temp.Label;
import temp.Temp;
import temp.TempMap;
import temp.TempList;

import frame.Frame;

import tree.MOVE;
import tree.BINOP;
import tree.MEM;
import tree.CONST;
import tree.SEQ;
import tree.TEMP;
import tree.Stm;
import tree.StmList;
import java.util.List;

public class SparcFrame extends Frame implements TempMap
{
    public final static String ENTRYPOINT = "main";
    public final static Registers registers = new Registers();
    public final static RegList localRegs() { return registers.localRegs(); }
    public final static RegList inRegs()    { return registers.inRegs(); }
    public final static RegList outRegs() { return registers.outRegs(); }
    public final static Temp    putRV() { return registers.putRV(); }
    public final static Temp    getRV() { return registers.getRV(); }

    // all the variables defined
    private final int WORD = 4;
    int nLocals;
    int nFormals;

    public Temp FP()   { return registers.fp(); }
    public Temp RV()   { return registers.putRV(); }
    public int frameSize() {
        int sz = WORD*(nLocals+nFormals);
        return -(96+sz) & -8;
    }
    public int wordSize()  { return WORD; }
    
    public String string(String string, Label label)
    {
        int len = string.length();
        StringBuffer output = new StringBuffer("\t.align 4\n" + label + ":");
        output.append("\t.asciz \"");
        output.append(string);
        output.append("\"\t\n");
        return output.toString();
    }

    public String label(String label) 
    { return "\t.align 4\n" + label + ":"; }
    
    public int allocLocal() {
        nLocals++;
        return (nLocals+nFormals)*-WORD; // offset
    }

    public int allocFormal() {
        nFormals++;
        return (nLocals+nFormals)*-WORD; // offset
    }

    public final List usableRegs() { return localRegs().toList(); }

    // add code to move from incoming registers to saved location
    public Stm procEntryExit1(Stm body) 
    {
        for(int i=0; i<nFormals; i++)
            body = new SEQ(new MOVE(new MEM(new BINOP(BINOP.PLUS,
                                                      new CONST(-4*(i+1)),
                                                      new TEMP(FP()))),
                                    new TEMP(regMap("%i"+i))),
                           body);
        return body;
    }

    public List procEntryExit3(List instr) 
    {
        int sz = -(96+frameSize()) & -8;
        sz = frameSize();
        instr.add(0, new assem.OPER("\tsave %sp, " + sz + ", %sp",
                                    null, null));
        instr.add(new assem.OPER("\tret", null, null));
        instr.add(new assem.OPER("\trestore\n", null, null));
        return instr;
    }

    // adds header and footer to code
    public Proc procEntryExit4(Proc proc) 
    {
        String prefix =
            "\t.global _start\n" +
            "\t.align 4\n";
        proc.prepend(prefix);

        proc.append("_start:");
        proc.append("\tcall\t" + ENTRYPOINT);
        proc.append("\tnop");
        proc.append("\tmov\t0, %o0	!success code");
        proc.append("\tmov\t1, %g1	!exit code");
        proc.append("\tta\t0");
        return proc;
    }
    
    public List codegen(StmList stms, regalloc.RegAlloc ra) 
    { return (new Codegen()).codegen(stms, ra); }

    public Temp regMap(String s)  { return registers.regMap(s); }
    public String tempMap(Temp t) { return registers.tempMap(t); }
}
