package symbol;

import java.util.HashMap;
import java.util.Stack;
import java.util.Vector;
import java.util.Iterator;
import symbol.*;

public class SymbolTable implements Table 
{
  private class Bucket {
    public Object value;
    public Bucket next;
        
    Bucket(Object value, Bucket next) {
      this.value = value;
      this.next = next;
    }
  }

  private class BucketMap extends HashMap {
    public void put(String key, Bucket bucket)
    { super.put(key, bucket); }

    public Bucket get(String key)
    { return (Bucket)super.get(key); }
  }  
  
  private BucketMap map = new BucketMap();
  private Stack scopes = new Stack();

  public boolean contains(String key) {
    return map.containsKey(key);
  }
    
  public void put(String key, Object value) {
    Vector scope = (Vector)scopes.peek();
    if(!scope.contains(key)) {
      scope.add(key);
      map.put(key, new Bucket(value, map.get(key)));
    }
    else
      // in order to prevent errors during redefines in the same scope
      map.put(key, new Bucket(value, map.get(key).next));
  }
    
  public Object get(String key) {
    Bucket bucket = map.get(key);
    return bucket==null ? null : bucket.value;
  }

  public void beginScope() {
    scopes.push(new Vector());
  }
    
  public void endScope() {
    Vector scope = (Vector)scopes.pop();
    Iterator iter = scope.iterator();
    while(iter.hasNext()) {
      String key = (String)iter.next();
      map.put(key, map.get(key).next);
    }
  }

  public Iterator keys() {
    return map.keySet().iterator();
  }

  /*
  public void print() 
  {
    // prints current symbols only for now
    Iterator iter = keys();
    while(iter.hasNext()) {
      
    }
  }
  */
}
