/*
  serves as an implementation for the frame
  because this compiler is only for a specific processor, for simplicity
  i'm skipping an unnecessary level of abstraction
*/

package translate;

import temp.Label;

import compiler.Compiler;
import frame.Frame;

import tree.Exp;
import tree.BINOP;
import tree.CONST;
import tree.MEM;
import tree.TEMP;

public class Level 
{
    public Frame frame;

    public Access staticLink;
    
    public AccessList locals  = new AccessList();
    public AccessList formals = new AccessList();    

    public Label label;  // identifier for the current frame
    public Level parent; // static-link to the parent frame

    public Level(Level parent)               { this(parent, new Label()); }
    public Level(Level parent, String label) { this(parent, new Label(label)); }
    public Level(Level parent, Label label) {
        frame = Compiler.newFrame();
        staticLink = allocFormal();
        this.label = label;
        this.parent = parent;
    }

    public int depth() {
        return (parent==null) ? 0 : 1+parent.depth();
    }
  
    public String toString() {
        return (parent==null) ? label.toString() : label.toString() + "/" + parent;
    }

    public Access allocLocal() {
        Access access = new Access(this, frame.allocLocal());
        locals.add(access);
        return access;
    }

    public Access allocFormal() {
        Access access = new Access(this, frame.allocFormal());
        formals.add(access);
        return access;
    }
    
    public String string(String string, Label label)
    { return frame.string(string, label); }

    /*
    // size of the frame, add 1 for the static-link
    public int frameSize() 
    { return frame.frameSize(); } 
    */
    
    /*
    public int offset(Level level) 
    {
        int offset = 0;
        Level curr = this;
        while(curr!=level) {
            curr = curr.parent;
            offset+=curr.frameSize();
        }
        return offset;
    }
    */
    
    // frame pointer
    public Exp getFP(Level level)
    {
        // details: the current fp is at %fp
        // current stack point is then at %fp + offset
        //   => parent fp is [%fp + offset]
        //   => next parent is [[%fp + offset] + offset]

        Exp exp = new TEMP(frame.FP());

        for(Level curr=this; curr!=level && curr!=null; curr=curr.parent) {
            exp = new MEM(new BINOP(BINOP.PLUS,
                                    exp,
                                    new CONST(staticLink.offset)));
        }
        return exp;
    }
}
