 /*
 * CSE3001.  Compiler project based on "Modern Compiler Implementation
 * in Java" by Andrew W. Appel.
 *
 */

package translate;

import error.ErrorMsg;
import parser.TigerParser;     // parser produced by CUP

import syntax.Exp;
import syntax.Print;

import semantics.Semantics;

import java_cup.runtime.Scanner;

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.IOException;

import java.util.LinkedList;
import java.util.Iterator;

public class Main {

   public static boolean verbose = System.getProperty ("verbose")!=null;

   public static void main (String args[]) throws IOException {
      final String filename = args[0];
      final ErrorMsg errorMsg = new ErrorMsg (filename);
      final InputStream is = new FileInputStream (filename);
      final TigerParser parser =  new TigerParser (is, errorMsg);
      final Print p = new Print (System.out);
      try {
	 final Exp e = (Exp) parser.parse().value;

         if(errorMsg.getErrors()>0) {
           System.out.println(errorMsg.getErrors() + " errors.");
           return;
         }
         
	 is.close();
	 if (verbose) {
	    p.prExp (e);
	 }
         final Semantics s = new Semantics(errorMsg);

         final LinkedList frags = s.transProg (e);
       
         for(Iterator i=frags.iterator(); i.hasNext(); )
           ((Fragment)i.next()).print(System.out);
      } catch(Exception e) {
        e.printStackTrace (System.err);
      }
      System.out.println(errorMsg.getErrors() + " errors.");
   }
}
